/* Sequential software instruction length decoder for IA32 
   Encoding bits as chars to make coding easy.
   Written by Shaun Xiushan Feng, May 31, 2005
*/

/* Reference: The IA-32 Intel Architecture Software Developer’s Manual,
   2004. Four volumes. Intel Order Numbers 253665–253668.
*/

#include <stdio.h>
#include <string.h>


#define PARCEL_SIZE 16          /* size of buffer           */
#define PAIR_SIZE  8            /* each "bitpair" is a byte */
#define false 		0
#define true		!(false)
#define INIT_RESERVED_OR_NOTDEFINED_FLAG false
#define INIT_ADDRESS_MODE true
#define INIT_OPERAND_MODE true



#define PreF0h			"11110000"
#define PreF2h			"11110010" 
#define Pre66h			"01100110"
#define Pre67h			"01100111"

/* one opcode byte instructions */
#define instr_AAA		"00110111"
#define instr_AAD		"11010101"
#define instr_AAM		"11010100" 
#define instr_AAS		"00111111"

#define instr_ADC_000100xx	"000100xx"
#define instr_ADC_100000x1      "100000x1"
#define instr_ADC_100000x0	"100000x0"
#define instr_ADC_00010101      "00010101"
#define instr_ADC_00010100	"00010100"


/* Opcode Extensions for one-and two-byte opcodes */


/* The IA-32 Intel Architecture Software Developer’s Manual,
   2004. Four volumes. Intel Order Numbers 253665–253668.
   Table A-4, OPCODE MAP, Intel Corporation.
*/
#define instr_opcode_extension_gr1_100000x1	"100000x1"
#define instr_opcode_extension_gr1_100000x0     "100000x0"

#define ESCAPE			"00001111"

/* two opcode bytes instructions */
/* start with x0Fh */
#define instr_BT_10100011 	"10100011"
#define instr_BT_10111010	"10111010"
#define instr_BSF               "10111100"
#define instr_BSWAP		"11001xxx"


/* global variables */

/* true: 32-bit addressing, false: 16-bit */
char address_mode_32; 
/* true: 32-bit operand,     false: 16-bit */
char operand_mode_32;   
/* if return with a true value, a reserved or nondefine pattern is found */
char reserved_or_notdefined_flag; 


void
error_flag_decoder()
{
        reserved_or_notdefined_flag = true;
        //exit(0);
}


void
init_decoder(
        char *wrapin)
{      
        int i, flag = 0;
 	reserved_or_notdefined_flag = INIT_RESERVED_OR_NOTDEFINED_FLAG;
  	address_mode_32 = INIT_ADDRESS_MODE;
  	operand_mode_32 = INIT_OPERAND_MODE;
        /* only one bit of wrapin can be 1 */
        for(i=0; i<PARCEL_SIZE; i++){      
          if (wrapin[i] == 1 && flag == 0) flag = 1;
          if (wrapin[i] == 1 && flag == 1) error_flag_decoder();
        }
}

/* I need to redefine a strcmp function such that 1111 \subset 11xx */
char
in_set(
        char *subset, 
	char *superset, 
	int  size)
{ 
  	int i;
        for (i=0; i<size; i++){
         if (subset[i]!=superset[i] &&
             superset[i] != 'x' && 
             superset[i] != 'X')
           return false;
        }
	return true;
}

void
read_one_PAIR(
        int  wrap,
        char *bitpair, 
        char *parcel,
        char *nextparcel)
{
        int i;
        if (wrap < PARCEL_SIZE)
          for (i=0; i<PAIR_SIZE; i++)
            bitpair[i] = parcel[wrap*PAIR_SIZE+i];
        else {
          for (i=0; i<PAIR_SIZE; i++)
            bitpair[i] = nextparcel[(wrap-PARCEL_SIZE)*PAIR_SIZE+i];
        }
}

/* Table 2-2 32-Bit Addressing Forms with ModR/M Byte */  
void
bit32_addressing_ModRM(
	int  *wrap,
	char *parcel, 
	char *nextparcel, 
	char *bitpair)
{
       /* Mod = 00 */
       if (in_set(bitpair,"00", 2) == true) {
           if (in_set(bitpair+5,"0",1) == true)
              return;
           if (in_set(bitpair+5,"11",2) == true)
              return;
           if (in_set(bitpair+5,"101",3) == true){
              (*wrap)+=4;
              return;
           }
         /* SIB byte follows */
           if (in_set(bitpair+5,"100",3) == true){
              read_one_PAIR(*wrap, bitpair, parcel, nextparcel);
              (*wrap)++;
              if (in_set(bitpair+2,"100",3) == false) {
                 if (in_set(bitpair+5,"101",3) == true) {
                    (*wrap)+=4;
                    return;
                 }
                 else
                    return;
           }
           else 
              /* notdefined */
              error_flag_decoder(); 
              return;
          }
        /* SIB byte is finished */
         return;
       }

       /* Mod = 01 */
       if (in_set(bitpair,"01", 2) == true) {
           if (in_set(bitpair+5,"100",3) == false){
             (*wrap)++;
             return;
           }
           /* SIB byte follows */
           if (in_set(bitpair+5,"100",3) == true){
              read_one_PAIR(*wrap, bitpair, parcel, nextparcel);
              (*wrap)++;
              if (in_set(bitpair+2,"100",3) == false) {
                 (*wrap)++;
                 return;
              }
              else
             /* notdefined */
                 error_flag_decoder();
              return;
           }
           /* SIB byte is finished */
        }   
       
       /* Mod = 10 */
       if (in_set(bitpair,"10", 2) == true) {
           if (in_set(bitpair+5,"100",3) == false){
              (*wrap)+= 4;
           return;
         }
         /* SIB byte follows */
         if (in_set(bitpair+5,"100",3) == true){
            read_one_PAIR(*wrap, bitpair, parcel, nextparcel);
            (*wrap)++;
            if (in_set(bitpair+2,"100",3) == false) {
                 (*wrap)+=4;
                 return;
            }
            else
             /* notdefined */
                 error_flag_decoder();
            return;
         }
           /* SIB byte is finished */
       }

       /* Mod = 11 */
       /* nothing follows ModR/M Byte. We do nothing for this case */
       /*
       if (in_set(bitpair,"11", 2) == true) {
       }
       */
}

/* Table 2-1 16-Bit Addressing Forms with ModR/M Byte */
void
bit16_addressing_ModRM(
        int  *wrap,
        char *bitpair)
{
       /* Mod = 00 */
       if (in_set(bitpair,"00", 2) == true) {
           if (in_set(bitpair+5,"110",3) == true)
               (*wrap)+=2;
           return;
       }
       /* Mod = 01 */
       if (in_set(bitpair,"01", 2) == true) {
           (*wrap)++;
           return;
       }
       /* Mod = 10 */
       if (in_set(bitpair,"10", 2) == true) {
           (*wrap)+=2;
           return;
       }
       /* Mod = 11 */
       /* nothing follows ModR/M Byte we do nothing for this case */
       /*
       if (in_set(bitpair,"11", 2) == true) {
       }
       */
}

/* bitpair is ModR/M byte */
void
mod_reg_rm(
	int *wrap, 
	char *parcel,
	char *nextparcel,
	char *bitpair)
{
	(*wrap)++; /* shift over ModR/M byte */
 	switch(address_mode_32){
     	   case true:
       		bit32_addressing_ModRM(wrap, parcel, nextparcel, bitpair);
      	        break;
  	   case false:
       		bit16_addressing_ModRM(wrap, bitpair); 
       	        break;
 	}  
}


/* I only handle prefix 0x66 and 0x67,
   which can change the instruction length */
char
handle_prefix(
	int *wrap,
	char *bitpair,
	char *parcel,
	char *nextparcel)
{
	int i;
        char pref_gr1, pref_gr2, pref_gr3, pref_gr4;
        pref_gr1 = false; pref_gr2 = false;
        pref_gr3 = false; pref_gr4 = false;

     /* Each instruction can have maximum 4 prefixes */
     for (i=0; i<4; i++) {
        read_one_PAIR(*wrap, bitpair, parcel, nextparcel);
        /* toggle with addressing or operand mode prefix */
        if (strcmp(bitpair,Pre66h) == 0) {
            if (pref_gr3 == false) {
           	operand_mode_32 = !(operand_mode_32);
                (*wrap)++;
                pref_gr3 =  true;
            }
            else /* error, only one prefix in each group */
            	error_flag_decoder();
            continue;
        }
        if (strcmp(bitpair,Pre67h) == 0) {
            if (pref_gr4 == false) {
            	address_mode_32 = !(address_mode_32);
            	(*wrap)++;
            	pref_gr4 =  true;
            }
            else 
                error_flag_decoder();
            continue;
        }

        /* Byte is not a prefix, bitpair has fresh data */
        if (!(pref_gr1 || pref_gr2 || pref_gr3 || pref_gr4))
          return false;
    }
    return true; /* Byte is a prefix, need to read another bitpair */
}

/* 
   Here is the part that you can play with 
   by adding new instructions, I put some sample instructions here.
*/

void
one_byte_opcode(
	int *wrap,
	char *parcel,
	char *nextparcel,
        char *bitpair,
	char *end)
{
	(*wrap)++;
	if (in_set(bitpair, instr_AAA, 8) == true) {
              if (*wrap-1<PARCEL_SIZE) end[*wrap-1]=1;
              return;
        }

	/* ADC r2r or m2r or r2m */
        if (in_set(bitpair, instr_ADC_000100xx, 8) == true){
               /* ModR/M Byte follows */
              read_one_PAIR(*wrap, bitpair, parcel, nextparcel);
       	      mod_reg_rm(wrap, parcel, nextparcel, bitpair);	 
	      if (*wrap-1<PARCEL_SIZE) end[*wrap-1]=1;	
	      return;
        }
 	/* ADC immediate to AL, AX or EAX */
        /* w = 1 */
        if (in_set(bitpair, instr_ADC_00010101, 8) == true){
               /* no ModR/M Byte follows, imm16 or imm32 follows */
              if (operand_mode_32 == true)
                  (*wrap)+=4; /* imm32 */
              else
                  (*wrap)+=2; /* imm16 */
              if (*wrap-1<PARCEL_SIZE) end[*wrap-1]=1;
              return;
        }
        /* w = 0 */
        if (in_set(bitpair, instr_ADC_00010100, 8) == true){
               /* no ModR/M Byte follows, w=0, imm8 follows */
              (*wrap)+=1;
              if (*wrap-1<PARCEL_SIZE) end[*wrap-1]=1;
              return;
        }

        /* notdefined */
        error_flag_decoder();

}

void
two_byte_opcode(
        int  *wrap,
	char *parcel,
	char *nextparcel,
        char *bitpair,
        char *end)
{
        (*wrap)++; /* shift over opcode, X0Fh has been shifted over */
	/* BSWAP */
        if (in_set(bitpair, instr_BSWAP, 8) == true){
              if (*wrap-1<PARCEL_SIZE) end[*wrap-1]=1;
              return;
        }

        /* notdefined */
        error_flag_decoder();
}

void
length_decoder(
        char    *parcel,        /* N "bitpair" parcel */
        char    *nextparcel,    /* N "bitpair" next parcel (needed lookahead) */
        char    *wrapin,        /* N "bit" input of wrap pointer */
        char    *begin,         /* N "bit" output of instruction starts. */
        char    *end,           /* N "bit" output of instruction ends. */
        char    *wrapout        /* N "bit" output of wrap pointer. */
)
{
        int     wrap;           /* wrap location as integer */
        char    bitpair[PAIR_SIZE];     /* */
        int     i;
	char 	ret;        

        for (i=0; i<PARCEL_SIZE; i++) {
                begin[i]=0;
                end[i]=0;
        }
        wrap = 0;
                                                                                
        /* First phase, skip over previously handled instructions */
        while (wrapin[wrap]==0)
                wrap++;
                                                                                
        if (wrap>0) end[wrap-1]=1;      /* end wrapped instruction */
                                                                                
        /* Next phase, start decoding instructions */
        while (wrap < PARCEL_SIZE){
	     begin[wrap]=1;

             operand_mode_32 = INIT_OPERAND_MODE;
             address_mode_32 = INIT_ADDRESS_MODE;

             ret = handle_prefix(&wrap,
                           bitpair,
                           parcel,
                           nextparcel);

	     if (ret == true)
                  read_one_PAIR(wrap, bitpair, parcel, nextparcel);
             /* match opcode */
             /* one byte opcode */
             if (in_set(bitpair, ESCAPE, 8) == false)
		  one_byte_opcode(&wrap, parcel, nextparcel, bitpair, end);
             
             /* two bytes opcode with escape x0Fh */ 
             if (in_set(bitpair, ESCAPE, 8) == true){
                  read_one_PAIR(wrap, bitpair, parcel, nextparcel);
	          two_byte_opcode(&wrap, parcel, nextparcel, bitpair, end);
	     }	
         }
         wrapout[wrap - PARCEL_SIZE] = 1;
}

int main(int argc, char *argv[])
{
        char    begin[PARCEL_SIZE],
                end[PARCEL_SIZE], 
                wrapin[PARCEL_SIZE], 
                wrapout[PARCEL_SIZE];
        char    parcel[PARCEL_SIZE * PAIR_SIZE],
                nextparcel[PARCEL_SIZE * PAIR_SIZE];
        int     i,j;
                                                                                                      
        for (i=0; i<PARCEL_SIZE; i++) wrapin[i]=0;
        wrapin[0] = 1;

        printf("Enter parcel as %d 0s or 1s:\n", PARCEL_SIZE * PAIR_SIZE);
        scanf("%s",parcel);
        for (i=0; i<PARCEL_SIZE * PAIR_SIZE; i++) parcel[i] -= '0';
        while (1) {
                printf("Enter next parcel as %d 0s or 1s:\n", PARCEL_SIZE * PAIR_SIZE);
                scanf("%s",nextparcel);
                for (i=0; i<PARCEL_SIZE * PAIR_SIZE; i++) nextparcel[i] -= '0';
                length_decoder(parcel,nextparcel,wrapin,begin,end,wrapout);

                printf("begin:\t");
                for (i=0; i<PARCEL_SIZE; i++) printf("%c  ",begin[i]+'0');
                printf("\n");

                printf("end:\t");
                for (i=0; i<PARCEL_SIZE; i++) printf("%c  ",end[i]+'0');
                printf("\n");

                printf("wrapout:\t");
                for (i=0; i<PARCEL_SIZE; i++) printf("%c  ",wrapout[i]+'0');
                printf("\n");
               
                for (i=0; i<PARCEL_SIZE * PAIR_SIZE; i++) parcel[i] = nextparcel[i];
                for (i=0; i<PARCEL_SIZE; i++) wrapin[i] = wrapout[i];
        }
}

